/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Alexander Schlemmer
 * Copyright (C) 2019-2020 IndiScale GmbH (info@indiscale.com)
 * Copyright (C) 2019-2020 Timm Fitschen (t.fitschen@indiscale.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */


/*
 * Check if an element is out of the viewport
 *
 * (c) 2018 Chris Ferdinandi, MIT License, https://gomakethings.com
 *
 * https://gomakethings.com/how-to-check-if-any-part-of-an-element-is-out-of-the-viewport-with-vanilla-js/
 *
 * @param {Node} elem - The element to check
 * @return {Object} A set of booleans for each side of the element.
 */
var isOutOfViewport = function (elem) {

    // Get element's bounding
    var bounding = elem.getBoundingClientRect();

    // Check if it's out of the viewport on each side
    var out = {};
    out.top = bounding.top < 0;
    out.left = bounding.left < 0;
    out.bottom = bounding.bottom > (window.innerHeight ||
        document.documentElement.clientHeight);
    out.right = bounding.right >
        (window.innerWidth || document.documentElement.clientWidth);
    out.any =
        out.top || out.left || out.bottom || out.right;
    out.all = out.top &&
        out.left && out.bottom && out.right;
    return out;
};


/**
 * @module awi_references
 * @version 0.1
 *
 * Special functionality for AWI. Should be removed from the main repository in
 * the future.
 *
 * @author Timm Fitschen
 */
var awi_references = new function () {

    var logger = log.getLogger("awi_references");


    this.find_bag_of_sample = async function (id) {
        return await
        this._find_ice_sample_back_ref(id, "Bag");
    }


    this.find_ice_core_of_sample = async function (id) {
        return await
        this._find_ice_sample_back_ref(id, "IceCore");
    }


    this._find_ice_sample_back_ref = async function (id, rt, oldcounter) {
        var counter = oldcounter + 1 || 1
        if (counter > 5) {
            return null;
        }
        var
            referencing_samples = await query(
                "FIND IceSample WHICH REFERENCES " +
                id);
        for (const sample of referencing_samples) {
            if (resolve_references.is_child(sample, rt)) {
                return sample;
            } else {
                var ret =
                    await
                this._find_ice_sample_back_ref(getEntityID(sample),
                    rt,
                    counter);
                if (ret) {
                    return ret;
                }
            }
        }
        return undefined;
    }

    const _stripe_re = /Stripe$/i;
    this.isStripe = function (el) {
        return _stripe_re.test(el.name)
    }


    this.get_icecore = async function (bag) {
        var id = getEntityID(bag);
        var icecore = (await query(
                "SELECT name FROM IceCore WHICH REFERENCES " +
                id))[0];
        var bag_number = getProperty(bag, "Number", false);
        var ret = {
                "data": {
                    "bag": bag_number
                }
            };
        if (!icecore) {
            ret["text"] =
                `${id} (Bag ${bag_number}, no Ice Core)`;
        } else {
            ret["text"] = `${id} (Ice Core ${getEntityName(icecore)}, Bag ${bag_number})`;
            ret["data"]["icecore"] = getEntityName(icecore);
        }
        return ret;
    }


    this.get_bag_and_icecore = async function (sample) {
        var id =
            getEntityID(sample);
        var bag = await awi_references.find_bag_of_sample(id);
        var ret = {};
        if (!bag) {
            var icecore = await awi_references.find_ice_core_of_sample(id);
            if (!icecore) {
                ret["text"] = `${id} (Sample w/o Bag or Ice Core)`;
            } else {
                ret["text"] = `${id} (Ice Core ${getEntityName(icecore)}, no Bag)`;
                ret["data"] = {
                    "icecore": getEntityName(icecore)
                };
            }
        } else {
            return await awi_references.get_icecore(bag);
        }
        return ret;
    }


    this.summarize_subsamples = function (ref_infos) {
        logger.trace("enter summarize_subsamples ", ref_infos);
        var icecores = {};
        for (const ref_info of ref_infos) {
            const icecore_name = ref_info.data.icecore || "none";
            if (!icecores[icecore_name]) {
                icecores[icecore_name] = [];
            }
            const bagnumber = parseInt(ref_info.data.bag, 10);
            icecores[icecore_name].push(bagnumber);
        }
        var ret = "";
        var last = "";
        const pretty_bag_numbers = reference_list_summary
            .simplify_integer_numbers;
        for (const icecore_name of
                Object.keys(icecores)) {
            if (icecore_name === "none") {
                last =
                    `<div class="casodb-f-resolve-reference-summary-plain">Bags without IceCore: ${pretty_bag_numbers(icecores[icecore_name])}</div>`;
            } else {
                ret +=
                    `<div class="caosdb-f-resolve-reference-summary-plain">IceCore: ${icecore_name} (Bags: ${pretty_bag_numbers(icecores[icecore_name])})</div>`;
            }
        }
        return ret.length + last.length > 0 ?
            '<b>Summary</b>' + ret + last : "";
    }


    this.summarize_box_content = function (ref_infos) {
        logger.trace("enter summarize_box_content ", ref_infos);
        return awi_references.summarize_subsamples(ref_infos);
    }
}

/**
 * @module reference_list_summary
 * @version 0.1
 *
 * For generating short summaries of LIST properties. This module is used by
 * the resolve_references module.
 *
 * @author Timm Fitschen
 */
var reference_list_summary = new function () {

    var logger = log.getLogger("reference_list_summary");

    /** Return a condensed string representation of an array of integers.
     *
     * @example simplify_integer_numbers([1,2,3,5,8,9,10]) returns "1-3, 5,
     * 8-10"
     *
     * @example simplify_integer_numbers([]) returns ""
     *
     * @example simplify_integer_numbers([1]) returns "1"
     *
     * @example simplify_integer_numbers([1,2]) returns "1, 2" because an array
     * with two elements gets a special treatment.
     *
     * The array may be unsorted.  simplify_integer_numbers([1,2,3]) returns
     * "1-3" simplify_integer_numbers([2,1,3]) returns "1-3" as well
     *
     * The array may contain duplicates.  simplify_integer_numbers([1,2,2,3])
     * returns "1-3"
     *
     * @param {numbers} array - unsorted array of integers, possibly with
     * duplicates.  @return {string} a condensed string representation of the
     * array.
     */
    this.simplify_integer_numbers = function (array) {
        logger.trace("enter simplify_integer_numbers", array);
        var set = Array.from(new Set(array));

        if (set.length === 0) {
            return ""
        } else if (set.length === 1) {
            return `${set[0]}`;
        }

        // sort numerically
        set.sort((a, b) => a - b);

        if (set.length === 2) {
            return `${set[0]}, ${set[1]}`;
        }


        var ret = `${set[0]}`;
        var last = undefined;
        // set[0];

        // e.g. [1,2,3,4,5,8,9,10];
        for (const next of set) {
            // append '-' to summarize consecutive numbers
            if (next - last === 1 && !ret.endsWith("-")) {
                ret += "-";
            }

            if (next - last > 1) {

                if (ret.endsWith("-")) {
                    // close previous interval and start new
                    ret += `${last}, ${next}`;
                } else {
                    // no previous interval, start interval.
                    ret += `, ${next}`;
                }
            } else if (next === set[set.length - 1]) {
                // finish interval if next is last item
                ret += next;
                break;
            }


            last = next;

        }

        // e.g. "1-5, 8-10"
        return ret;
    }

    /**
     * Generate a summary of all reference_infos by calling the callback
     * function of the first reference_info in the array.
     *
     * The summary is appended to the summary_container if available and
     * returned (for testing purposes).
     *
     * Returns undefined, if ref_infos is empty or does not come with a
     * callback function.
     *
     * @param {reference_info[]} ref_infos - array of reference_infos.
     * @param {HTMLElement} summary_container - the summary is appended to this
     *     element.
     * @return {HTMLElement|string} generated summary
     */
    this.generate = function (ref_infos, summary_container) {
        logger.trace("enter generate", ref_infos);
        if (ref_infos.length > 0 &&
            typeof ref_infos[0].callback === "function") {
            const summary =
                ref_infos[0].callback(ref_infos);
            if (summary && summary_container) {
                $(summary_container).append(summary);
            }
            logger.trace("leave generate", summary);
            return summary;
        }
        logger.trace("leave generate, return undefined");
        return undefined;
    }
}


/**
 * @module resolve_references
 * @version 0.2
 *
 * Original implementation by Alexander Schlemmer, 11/2018
 *
 * @author Timm Fitschen
 * @author Alexander Schlemmer
 */
var resolve_references = new function () {

    var logger = log.getLogger("resolve_references");

    var _scroll_timeout = undefined;

    /**
     * Scroll listener calls {@link resolve_references.update_visible_references} 500 milliseconds after the
     * last scroll event.
     */
    var scroll_listener = () => {
        if (_scroll_timeout) {
            clearTimeout(_scroll_timeout);
        }
        _scroll_timeout = setTimeout(function () {
            resolve_references.update_visible_references();
        }, 500);
    };


    /**
     * Initilize the scroll listener which triggers the resolution of the
     * entity ids and trigger it for the first time in order to resolve all
     * visible references.
     */
    this.init = function () {
        scroll_listener();

        // mainly for vertical scrolling
        $(window).scroll(scroll_listener);

        // for horizontal scrolling. is this still necessary when lists are
        // loaded in one go?
        $(".caosdb-value-list").scroll(scroll_listener);
    }

    /**
     * Check if an element is inside of the viewport on the vertical axis.
     *
     * Returns true if any of the element's bounding box's top or bottom edges are
     * in the viewport.
     *
     * @param {HTMLElement} elem - the element to check @return {boolean}
     *
     */
    this.is_in_viewport_vertically = function (elem) {
        var out =
            isOutOfViewport(elem);
        return !(out.top || out.bottom);
    }

    /** Check if an element is inside of the viewport on the horizontal axis.
     *
     * This criterion is very much different from the vertical correspondent: It
     * looks for the parent which contains the list scroll bar for caosdb values.
     * Then it is checked whether the element's bounding box is visible inside the
     * scroll box.
     *
     * @param {HTMLElement} elem - the element to check @return {boolean}
     *
     */
    this.is_in_viewport_horizontally = function (elem) {
        var scrollbox = elem.parentElement.parentElement;
        // Check this condition only if the grand parent is a list and return true
        // otherwise.
        if (scrollbox.classList.contains("caosdb-value-list") ==
            true) {
            var boundel = elem.getBoundingClientRect();
            var boundscroll = scrollbox.getBoundingClientRect();
            var leftcrit = boundel.right > boundscroll.left;
            var rightcrit = boundel.left < boundscroll.right;
            return leftcrit && rightcrit;
        } else {
            return true;
        }
    }


    /**
     * Return the name of a person as firstname + lastname
     */
    this.get_person_str = function (el) {
        var valpr = getProperties(el);
        if (valpr == undefined) {
            return;
        }
        return valpr.filter(valprel =>
                valprel.name.toLowerCase() == "firstname")[0].value +
            " " +
            valpr.filter(valprel => valprel.name.toLowerCase() ==
                "lastname")[0].value;
    }


    /**
     * Return true iff the entity has at least one direct parent named `par`.
     *
     * @param {HTMLElement} entity - entity in HTML representation.  @param
     * {string} par - parent name.  @return {boolean}
     */
    this.is_child = function (entity, par) {
        var pars = getParents(entity);
        for (const par of pars) {
            if (par.name === par) {
                return true;
            }
        }
        return false;
    }


    /**
     * Example implementation of a function which returns a reference_info for
     * referenced `ReferenceObject` entities.  `ReferenceObject` is a mock-up
     * entity which is used to test this module.
     */
    this.get_referenced = function (entity) {
        return {
            "text": getEntityName(entity),
            "data": {
                "name": getEntityName(entity),
                "number": getEntityName(entity).replace(
                    "ReferenceObject-", "")
            },
            "callback": function (ref_infos) {
                var ret = $('<div/>').append("Summary: ");
                var
                    array = []
                for (const ref_info of ref_infos) {
                    array.push(parseInt(ref_info["data"]["number"],
                        10));
                }
                ret.append(reference_list_summary
                    .simplify_integer_numbers(array));
                return ret[0];
            }
        };
    }


    this.retrieve = retrieve;

    /**
     * @typedef {reference_info}
     * @property {string} text
     * @property {function} callback - a callback function with one parameter
     *     (`data`) which generates a summary from the array of data objects
     *     which are passed to this function.
     * @property {object} data - an object with properties which can be
     *     understood and used by the `callback` function.
     */

    /**
     * Return a reference_info for an entity.
     *
     * TODO refactor to be configurable.  @async @param {string} id - the id of
     * the entity which is to be resolved.  @return {reference_info}
     */
    this.resolve_reference = async function (id) {
        const entity = (await resolve_references.retrieve(id))[0];

        // TODO handle multiple parents
        const par = getParents(entity)[0] || {};

        var ret = {
            "text": id
        };
        if (getEntityHeadingAttribute(entity, "path") !==
            undefined || par.name == "Image") {
            // show file name
            var pths = getEntityHeadingAttribute(entity, "path")
                .split("/");
            ret["text"] = pths[pths.length - 1];
        } else if (par.name === "Person") {
            ret["text"] = this.get_person_str(entity);
        } else if (par.name === "ExperimentSeries") {
            ret["text"] =
                getEntityName(entity);
        } else if (par.name === "BoxType") {
            ret["text"] = getEntityName(entity);
        } else if (par.name === "Loan") {
            var borrower = await this.retrieve(getProperty(entity, "Borrower"));
            var loan_state = awi_demo.get_loan_state_string(getProperties(entity));
            ret["text"] = "Borrowed by " + this.get_person_str(borrower[0]) + " (" + loan_state.replace("_", " ") + ")";
        } else if (par.name === "SubSample" || par.name === "BagMean" || awi_references.isStripe(par)) {
            ret = await awi_references.get_bag_and_icecore(entity);
            ret["callback"] = awi_references.summarize_subsamples;
        } else if (par.name === "Bag") {
            ret = await awi_references.get_icecore(entity);
            ret["callback"] = awi_references.summarize_box_content;
        } else if (par.name === "Box") {
            ret["text"] = getProperty(entity, "Number");
        } else if (par.name === "Palette") {
            ret["text"] = getProperty(entity, "Number"); } else if (par.name === "Referenced") {
            ret = this.get_referenced(entity);
        } else {
            var name = getEntityName(entity);
            if (typeof name !== "undefined" && name.length > 0) {
                ret["text"] = name;
            }
        }

        return ret;
    }


    this._target_class = "caosdb-resolve-reference-target";

    /**
     * Add a target span where the resolved reference information can be shown.
     *
     * If the element has a target yet, the existing one is returned.
     *
     * @param {HTMLElement} element - where to append the target.
     * @return {HTMLElement} the new/existing target element.
     */
    this.add_target = function (element) {
        if(element.getElementsByClassName(this._target_class).length > 0){
            return element.getElementsByClassName(this._target_class);
        } else {
            return $(`<span class="${this._target_class}"/>`)
                .appendTo(element)[0];
        }
    }

    /**
     * Retrieve meaningful information for a single caosdb-f-reference-value
     * element.
     *
     * @param {HTMLElement} rs - resolvable reference link
     * @return {reference_info} the resolved reference information
     */
    this.update_single_resolvable_reference = async function (rs) {
        const target = this.add_target(rs);
        const id = getIDfromHREF(rs);
        target.textContent = id;
        const resolved_entity_info = await this.resolve_reference(id);
        target.textContent = resolved_entity_info.text;
        return resolved_entity_info;
    }


    /**
     * Add a summary field to the the list_values element.
     *
     * A summary field is a DIV element with class
     * `caosdb-resolve-reference-summary`. The summary field is used to display
     * a condensed string representation of the referenced entities in the
     * list-property.
     *
     * @param {HTMLElement} list_values - where to add the summary field.
     * @return {HTMLElement} a summary field.
     */
    this.add_summary_field = function (list_values) {
        const summary = $(
            '<div class="caosdb-resolve-reference-summary"/>');
        $(list_values).prepend(summary);
        return summary[0];
    }

    this._unresolved_class_name = "caosdb-resolvable-reference";

    this.get_resolvable_properties = function (container) {
        const _magic_class_name = this._unresolved_class_name;
        return $(container).find(".caosdb-f-property-value").has(
            `.${_magic_class_name}`).toArray();
    }


    /*
     * This function updates all references in the body which are inside of the
     * current viewport.
     *
     * If the optional container parameter is given, only elements inside the
     * container are being processed.
     *
     * @param {HTMLElement} container
     * @return {Promise[]} array of promises for ref_infos.
     */
    this.update_visible_references = function (container) {
        const property_values = resolve_references
            .get_resolvable_properties(container || document.body);

        const _magic_class_name = resolve_references
            ._unresolved_class_name;

        var all_ref_infos = [];
        for (const property_value of property_values) {
            const lists = $(property_value).find(
                ".caosdb-value-list").has(
                `.${_magic_class_name}`);

            if (lists.length > 0) {
                const summary_field = resolve_references
                    .add_summary_field(property_value);

                logger.debug("processing lists of references", lists);
                for (var i = 0; i < lists.length; i++) {
                    const list = lists[i];
                    if (resolve_references
                            .is_in_viewport_vertically(list)) {
                        const rs = $(list).find(
                                `.${_magic_class_name}`)
                            .toggleClass(_magic_class_name, false);
                        const ref_infos = [];
                        for (var j = 0; j < rs.length; j++) {
                            const ref_info = resolve_references
                                .update_single_resolvable_reference(rs[j]);
                            ref_info["index"] = j;
                            ref_infos.push(ref_info);
                        }

                        // wait for resolution of references,
                        // then generate the summary.
                        Promise.all(ref_infos)
                            .then(_ref_infos => reference_list_summary
                                .generate(_ref_infos, summary_field))
                            .catch(logger.error);

                        all_ref_infos = all_ref_infos.concat(ref_infos);
                    }
                }
            } else {
                // TODO merge code with the above?
                const rs = $(property_value).find(
                    `.${_magic_class_name}`);
                logger.debug("processing single references", rs);
                for (var i = 0; i < rs.length; i++) {
                    if (resolve_references.is_in_viewport_vertically(
                            rs[i]) &&
                        resolve_references.is_in_viewport_horizontally(
                            rs[i])) {
                        $(rs).toggleClass(_magic_class_name, false);
                        all_ref_infos.push(resolve_references
                            .update_single_resolvable_reference(rs[i]));
                    }
                }
            }
        }

        return all_ref_infos;
    }
}


$(document).ready(function () {
    caosdb_modules.register(resolve_references);
});
