/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2020 IndiScale GmbH <info@indiscale.com>
 * Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */

'use strict';

/**
 * The ext_revisions module extends the edit_mode update functionality.
 *
 * The edit_mode.update_entity function is overridden by this module with a
 * proxy pattern. That means, that the original function is still called, but a
 * proxy (or wrapper) function adds further functionality.
 *
 * The extended update function creates a back-up version of the updated entity
 * and adds a revisionOf property to the updated entity which references
 * the back-up. The back-up entity loses all of its original parents and gets
 * an "Obsolete" as only parent instead.
 *
 * Per default, the module assumes two Entities to be present in the
 * database. A RecordType named "Obsolete" and a Property named
 * "revisionOf". The initialization is aborted if these entities cannot be
 * found and the module remains inactive.
 *
 * @module ext_revisions
 * @version 0.1
 *
 * @requires jQuery
 * @requires log
 * @requires edit_mode
 * @requires getEntityID
 * @requires transaction
 * @requires _createDocument
 */
var ext_revisions = function ($, logger, edit_mode, getEntityID, transaction, _createDocument) {


    /**
     * Default names for the two entities which are required by this module.
     */
    var _datamodel = { obsolete: "Obsolete", revisionOf: "revisionOf" };

    /**
     * Generate and insert the back-up entity which stores the old state of the
     * entity which is to be updated.
     *
     * The obsolete entity has only one parent named "Obsolete".
     * Apart from that, the obsolete entity has all the properties, name,
     * description and so on from the original entity before the update.
     *
     * @param {string} id - the id of the entity which is to be updated.
     * @returns {string} the id of the newly created obsolete entity.
     */
    var _insert_obsolete = async function (id) {
        logger.debug("insert obsolete", id);

        // create new obsolete entity from the original
        const obsolete = await transaction.retrieveEntityById(id);
        $(obsolete).attr("id", "-1");
        $(obsolete).find("Permissions").remove();
        $(obsolete).find("Parent").remove();
        $(obsolete).append(`<Parent name="${_datamodel.obsolete}"/>`);

        const doc = _createDocument("Request");
        doc.firstElementChild.appendChild(obsolete);
        const result = await transaction.insertEntitiesXml(doc);
        const obsolete_id = $(result.firstElementChild).find("[id]").first().attr("id");
        logger.trace("leave _insert_obsolete", obsolete_id);
        return obsolete_id;
    };

    /**
     * Generate a HTML string which represents a new "revisionOf" property
     * which references the newly created obsolete entity. The property is
     * meant to be appended to the property section of the entity which is to
     * be updated.
     *
     * @param {string} obsolete_id - the id of the newly created obsolete
     *     entity.
     * @returns {string} A HTML represesentation of an entity property.
     */
    var _make_revision_of_property = async function (obsolete_id) {
        logger.trace("enter _make_revision_of_property", obsolete_id);
        const ret = (await transformation.transformProperty(str2xml(`<Response><Property id="${_datamodel._revisionOfId}" name="${_datamodel.revisionOf}" datatype="${_datamodel.obsolete}"></Property></Response>`))).firstElementChild;

        $(ret).find(".caosdb-f-property-value").append(`<div class="caosdb-property-edit-value"><select><option value="${obsolete_id}" selected="selected"></option></select></div>`);

        logger.trace("leave _make_revision_of_property", ret);
        return ret;
    }

    /**
     * Remove all properties from ent_element with the id of the "revisionOf"
     * property.
     *
     * @param {HTMLElement} ent_element - entity in HTML representation.
     */
    var _remove_old_revision_of_property = function (ent_element) {
        $(ent_element)
            .find(".caosdb-f-entity-property")
            .filter(function(index, property) {
                if(_datamodel._revisionOfId === $(property)
                    .find(".caosdb-property-id").text()) {
                    return true;
                }
                return false;
        }).remove();
    }

    /**
     * Main functionality of this module is in here.
     *
     * This method is called by the (overridden) edit_mode.update_entity
     * function before the actual update. It inserts a new obsolete entity
     * which represesents the old state of the entity, deletes any revisionOf
     * properties of the entity (if present) and adds a new revisionOf property
     * which references the (newly inserted) obsolete entity.
     *
     * @param {HTMLElement} ent_element - The entity form which has been
     * generated by the edit_mode with the changes by the user.
     */
    var _create_revision = async function (ent_element) {
        logger.debug("create revision", ent_element);
        var id = getEntityID(ent_element);

        var obsolete_id = await _insert_obsolete(id);

        // remove old revision of and add new one
        _remove_old_revision_of_property(ent_element);
        var revision_of_property = await _make_revision_of_property(obsolete_id);
        var properties_section = ent_element.getElementsByClassName("caosdb-properties")[0];
        properties_section.appendChild(revision_of_property);
    };

    /**
     * Test whether the necessary entities exist ("revisionOf" and "Obsolete").
     */
    var _check_datamodel = async function() {
        var results = Promise.all([
            query(`FIND RecordType ${_datamodel.obsolete}`),
            query(`FIND Property ${_datamodel.revisionOf}`)
        ]);

        for (let result of (await results)) {
            if (result.length !== 1) {
                throw new Error("Invalid datamodel");
            }

            var name = getEntityName(result[0]);
            if (name && name.toLowerCase() === _datamodel.revisionOf.toLowerCase()) {
                _datamodel._revisionOfId = getEntityID(result[0]);
                _datamodel.revisionOf = name;
            } else if (name && name.toLowerCase() === _datamodel.obsolete.toLowerCase()) {
                _datamodel.obsolete = name;
            }
        }
    };

    /**
     * Initialize the ext_revisions module.
     *
     * Per default, the module assumes two Entities to be present in the
     * database. A RecordType named "Obsolete" and a Property named
     * "revisionOf". The initialization is aborted if these entities cannot be
     * found and the module remains inactive. For testing purposes the names of
     * these entities can be set to different values via the respective
     * parameters.
     *
     * @param {string} [obsolete] - The name of the obsolete RecordType.
     * @param {string} [revisionOf] - The name of the revisionOf Property.
     */
    var init = async function (obsolete, revisionOf) {
        if (typeof obsolete === "string") {
            _datamodel.obsolete = obsolete;
        }
        if (typeof revisionOf === "string") {
            _datamodel.revisionOf = revisionOf;
        }

        try {
            await _check_datamodel();
        } catch (err) {
            logger.error("could not init ext_revisions", err);
            return;
        }

        (function(proxied) {
            edit_mode.update_entity = async function(ent_element) {
                await _create_revision(ent_element);
                return await proxied.apply(this, arguments)
            };
        })(edit_mode.update_entity);
    }

    return {
        // public members, part of the API
        init: init,
        // private members, exposed for testing
        _make_revision_of_property: _make_revision_of_property,
        _datamodel: _datamodel,
        _logger: logger,
    }
}($, log.getLogger("ext_revisions"), edit_mode, getEntityID, transaction, _createDocument);


// this will be replaced by require.js in the future.
$(document).ready(function () {
    if ("${BUILD_MODULE_EXT_REVISIONS}" == "ENABLED") {
        caosdb_modules.register(ext_revisions);
    }
});
