/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2019 IndiScale GmbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */

'use strict';

QUnit.module("ext_map.js", {
    before: function(assert) {
        var lat = "latitude";
        var lng = "longitude";
        this.datamodel = { lat: lat, lng: lng };
        this.test_map_entity = `
<div class="caosdb-entity-panel caosdb-properties">
  <div class="caosdb-id">1234</div>
  <div class="list-group-item caosdb-f-entity-property">
    <div class="caosdb-property-name">` +
            lat + `</div>
    <div class="caosdb-f-property-value">1.23</div>
  </div>
  <div class="list-group-item caosdb-f-entity-property">
    <div class="caosdb-property-name">` +
            lng + `</div>
    <div class="caosdb-f-property-value">5.23</div>
  </div>
</div>`;
    },
    beforeEach: function(assert) {
        sessionStorage.removeItem("caosdb_map.view");
    }
});

QUnit.test("availability", function(assert) {
    assert.equal(caosdb_map.version, "0.3", "test version");
    assert.ok(caosdb_map.init, "init available");
});

QUnit.test("default config", function(assert) {
    assert.ok(caosdb_map._default_config);
    assert.equal(caosdb_map._default_config.version, caosdb_map.version, "version");
});

QUnit.test("load_config", async function(assert) {
    assert.ok(caosdb_map.load_config, "available");
    var config = await caosdb_map.load_config("non_existing.json");
    assert.ok(config, "returns something");
    assert.equal(config.views.length, 1, "one view in default");
    assert.equal(config.views[0].id, "UNCONFIGURED", "view has id 'UNCONFIGURED'.");
});

QUnit.test("check_config", function(assert) {
    assert.ok(caosdb_map.check_config(caosdb_map._default_config), "default config ok");
    assert.throws(()=>caosdb_map.check_config({"version": "wrong version",}), /The version of the configuration does not match the version of this implementation of the caosdb_map module. Should be '.*', was 'wrong version'./, "wrong version");
});

QUnit.test("check dependencies", function(assert) {
    assert.ok(caosdb_map.check_dependencies, "available");
    assert.propEqual(caosdb_map.dependencies, ["log", {"L": ["latlngGraticule", "Proj"]}, "navbar", "caosdb_utils"]);
    assert.ok(caosdb_map.check_dependencies(), "deps available");
});

QUnit.test("create_toggle_map_button", function(assert) {
    assert.ok(caosdb_map.create_toggle_map_button, "available");
    var button = caosdb_map.create_toggle_map_button();
    assert.equal(button.tagName, "BUTTON", "is button");
    assert.ok($(button).hasClass("caosdb-f-toggle-map-button"),"has caosdb-f-toggle-map-button class");
    assert.equal($(button).text(), "Map", "button says 'Map'");

    // set other content:
    button = caosdb_map.create_toggle_map_button("Karte");
    assert.equal(button.tagName, "BUTTON", "is button");
    assert.ok($(button).hasClass("caosdb-f-toggle-map-button"),"has caosdb-f-toggle-map-button class");
    assert.equal($(button).text(), "Karte", "button says 'Karte'");

});

QUnit.test("bind_toggle_map", function(assert) {
    let button = $("<button/>")[0];
    let done = assert.async();

    assert.ok(caosdb_map.bind_toggle_map, "available");
    assert.throws(()=>caosdb_map.bind_toggle_map(button), /parameter 'toggle_cb'.* was undefined/, "no function throws");
    assert.throws(()=>caosdb_map.bind_toggle_map("test", ()=>{}), /parameter 'button'.* was string/, "string button throws");
    assert.equal(caosdb_map.bind_toggle_map(button, done), button, "call returns button");

    // button click calls 'done'
    $(button).click();
});


QUnit.test("create_map", function(assert) {
    assert.equal(typeof caosdb_map.create_map_view, "function", "function available");

});

QUnit.test("create_map_panel", function(assert) {
    assert.ok(caosdb_map.create_map_panel, "available");
    let panel = caosdb_map.create_map_panel();
    assert.equal(panel.tagName, "DIV", "is div");
    assert.ok($(panel).hasClass("caosdb-f-map-panel"), "has class caosdb-f-map-panel");
    assert.ok($(panel).hasClass("container"), "has class container");
});

QUnit.test("create_map_view", function(assert) {
    var view_config = $.extend(true, {}, caosdb_map._unconfigured_views[0],
        {"select": true, "view_change": true});
    var map_panel = $("<div/>");

    var map = caosdb_map.create_map_view(map_panel[0], view_config);

    console.log(map_panel[0]);
    assert.ok(map instanceof L.Map, "map instance created");
    assert.ok(map_panel.hasClass("leaflet-container"), "map_panel has .leaflet-container child");

    assert.notOk(map._crs, "no special crs");
    map.remove();

    // test with pre-defined crs
    view_config["crs"] = "Simple";
    map = caosdb_map.create_map_view(map_panel[0], view_config);

    console.log(map_panel[0]);
    assert.ok(map instanceof L.Map, "map instance created");
    assert.ok(map_panel.hasClass("leaflet-container"), "map_panel has .leaflet-container child");
    assert.equal(map._crs, L.CRS.Simple, "map has SIMPLE crs");

    map.remove();

    // test with special crs:
    view_config["crs"] =  {
        "code": "EPSG:3995",
        "proj4def": "+proj=stere +lat_0=90 +lat_ts=71 +lon_0=0 +k=1 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs",
        "options": {
            "resolutions": ["16384", "8192", "4096", "2048", "1024", "512", "256", "128", "64", "32", "16", "8", "4", "2", "1", "0.5"]
        }
    };

    map = caosdb_map.create_map_view(map_panel[0], view_config);

    console.log(map_panel[0]);
    assert.ok(map instanceof L.Map, "map instance created");
    assert.ok(map_panel.hasClass("leaflet-container"), "map_panel has .leaflet-container child");
    assert.ok(map._crs instanceof L.Proj.CRS, "map has special crs");

    map.remove();

});

QUnit.test("get_map_entities", function(assert) {
    var datamodel = this.datamodel;
    var container = $('<div/>').append(this.test_map_entity);
    var map_objects = caosdb_map.get_map_entities(container[0], datamodel);
    assert.equal(map_objects.length, 1, "has one map object");
});


QUnit.test("create_entitiy_markers", function(assert) {
    var datamodel = this.datamodel;
    var entities = $(this.test_map_entity).toArray();

    // w/o popup
    var markers = caosdb_map.create_entitiy_markers(entities, datamodel);
    assert.equal(markers.length, 1, "has one marker");
    assert.ok(markers[0] instanceof L.Marker, "is marker");
    var latlng = markers[0]._latlng;
    assert.equal(latlng.lat, "1.23", "latitude set");
    assert.equal(latlng.lng, "5.23", "longitude set");
    assert.notOk(markers[0].getPopup(), "no popup");

    // with popup
    var markers = caosdb_map.create_entitiy_markers(entities, datamodel, ()=>"popup");
    assert.ok(markers[0].getPopup(), "has popup");
});


QUnit.test("_add_current_page_entities", function(assert) {
    var datamodel = this.datamodel;
    var layerGroup = L.layerGroup();
    var container = $('<div class="caosdb-f-main-entities"/>').append(this.test_map_entity);
    $("body").append(container);

    assert.equal(layerGroup.getLayers().length, 0, "no layer");
    var cpe = caosdb_map._get_current_page_entities(datamodel, undefined, undefined, undefined, undefined);

    assert.equal(cpe.length, 1, "has one entity");
    container.remove();
});


QUnit.test("make_layer_chooser_html", function(assert) {
    var test_conf = { "id": "test_id",
        "name": "test name",
        "description": "test description",
        "icon": { "html": "<span>ICON</span>",
        },
    };

    var layer_chooser = caosdb_map.make_layer_chooser_html(test_conf);
    assert.ok(layer_chooser, "available");
    assert.equal($(layer_chooser).attr("title"), "test description", "description set as title");
});

QUnit.test("init_entity_layer", function(assert) {
    var done = assert.async();
    var test_conf = { "id": "test_id",
        "name": "test name",
        "description": "test description",
        "get_entities": async function() {done(); return []},
        "icon": { "html": "<span>ICON</span>",
        },
    }

    var entityLayer= caosdb_map.init_entity_layer(test_conf);
    assert.equal(entityLayer.id, test_conf.id, "id");
    assert.equal(entityLayer.active, true, "is active");
    assert.ok(entityLayer.chooser_html instanceof HTMLElement, "chooser_html is HTMLElement");
    assert.equal(entityLayer.layer_group.getLayers().length, 0 , "empty layergroup");
});
