/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
/* testing webcaosdb's javascript sources */

/* SETUP ext_references module */
QUnit.module("ext_references.js", {
    before: function(assert) {
        // dummy test functions
        var logger = log.getLogger("ext_references.js")
        logger.setLevel("debug");

        resolve_references.retrieve = async function (id) {
            logger.debug("retrieve", id);
            return transformation.transformEntities(str2xml(
                `<Response>
                  <Record id="${id}" name="ReferenceObject-${id}">
                    <Parent name="Referenced"/>
                  </Record>
                </Response>`
            ));
        }

        resolve_references.is_in_viewport_horizontally = () => true;
        resolve_references.is_in_viewport_vertically = () => true;
    }
});

QUnit.test("available", function(assert) {
    assert.ok(resolve_references);
});

QUnit.test("init", function(assert){
    assert.ok(resolve_references.init);
});

QUnit.test("get_person_str", function(assert){
    assert.ok(resolve_references.get_person_str);
});

QUnit.test("update_visible_references", async function(assert){
    const f = resolve_references.update_visible_references;

    const test_property = $(`<div class="caosdb-f-property-value">
        <div data-entity-id="15"
            class="${resolve_references._unresolved_class_name}">
          <span class="${resolve_references._target_class}"/></span>
        </div>
    </div><div class="caosdb-f-property-value">
        <div class="caosdb-value-list">
            <div data-entity-id="16"
                class="${resolve_references._unresolved_class_name}">
              <span class="${resolve_references._target_class}"/></span>
            </div>
            <div data-entity-id="17"
                class="${resolve_references._unresolved_class_name}">
              <span class="${resolve_references._target_class}"/></span>
            </div>
        </div>
    </div>`);

    assert.equal(test_property.find(`.${resolve_references._unresolved_class_name}`).length, 3, "is unresolved");

    const ref_infos = f($("<div/>").append(test_property));

    assert.equal(test_property.find(`.${resolve_references._unresolved_class_name}`).length, 0, "is resolved");


    assert.equal(ref_infos.length, 3, "three ref_infos");
    var done = assert.async();
    Promise.all(ref_infos).then(ref_infos => {
        assert.equal(ref_infos[1].data.name.replace(/[0-9]*/g, ""), "ReferenceObject-", "has data");
        assert.ok(typeof ref_infos[1].callback === "function",
            "has callback");
        done();
    });

});


/*
 * This test checks whether all required
 * HTML elements are present.
 */
QUnit.test("check_structure_html", function(assert){
    var done = assert.async();
    let string_test_document = `
<Response>
<Record description="This record has no name.">
    <Parent name="bla" />

    <Property name="A" datatype="INTEGER">245</Property>

    <Property name="A" datatype="Uboot">245</Property>

    <Property name="A" datatype="LIST&#60;Uboot&#62;">
      <Value>245</Value>
      <Value>245</Value>
    </Property>
  </Record>
</Response>
`;
    var xmltestdocument = str2xml(string_test_document);
    transformation.transformEntities(xmltestdocument).then(x => {
        console.log(x);

        // Check that there are only three resolvable references:
        var elms = x[0].getElementsByClassName("caosdb-resolvable-reference");
        assert.equal(elms.length, 3);

        // Check that each of the elements of the list has a grand parent with a specific class:
        for (var i=1; i<elms.length; i++) {
            assert.equal(elms[i].parentElement.parentElement.classList.contains("caosdb-value-list"), true);
        }
        // but the first element does NOT:
        assert.equal(elms[0].parentElement.parentElement.classList.contains("caosdb-value-list"), false);

        done();
    }, err => {console.log(err);});
});

QUnit.module("reference_list_summary");

QUnit.test("simplify_integer_numbers", function(assert) {
    var f = reference_list_summary.simplify_integer_numbers;
    assert.equal(f([1]), "1");
    assert.equal(f([1,1,1]), "1");
    assert.equal(f([1,2]), "1, 2");
    assert.equal(f([2,1]), "1, 2");
    assert.equal(f([2,1,2]), "1, 2");
    assert.equal(f([2,1,2]), "1, 2");
    assert.equal(f([1,2,4]), "1-2, 4");
    assert.equal(f([1,3,4,5]), "1, 3-5");
    assert.equal(f([1,2,3,5,8,9,10]), "1-3, 5, 8-10");
    assert.equal(f([1,2,3,5,7,8,10]), "1-3, 5, 7-8, 10");
    assert.equal(f([1,3,5,7,8,10]), "1, 3, 5, 7-8, 10");
    assert.equal(f([1,2,4,5,7,8,10]), "1-2, 4-5, 7-8, 10");
    assert.equal(f([1,2,4,5,7,9,10]), "1-2, 4-5, 7, 9-10");
});


QUnit.test("generate", function(assert) {
    const f = reference_list_summary.generate;
    const summary_container = $("<div/>");


    // without callback function
    const ref_infos_no_cb = [ {
        "data": { "dummy": "dummy_data1" },
    }, {
        "data": { "dummy": "dummy_data2" },
    }, ];

    assert.ok(typeof f(ref_infos_no_cb) === "undefined", "no cb returns undefined");
    f(ref_infos_no_cb, summary_container);
    assert.equal(summary_container.children().length, 0, "no cb doesn't append to summary container.");
    assert.equal(summary_container.text(), "", "no text in container");

    // with callback function
    const ref_infos_with_cb = [ {
        "callback": function (ref_infos) { 
            return ref_infos.map(i => i.data.dummy).join(",");
        },
        "data": { "dummy": "dummy_data3" },
    }, {
        "data": { "dummy": "dummy_data4" },
    }, ];

    const summary = "dummy_data3,dummy_data4";

    assert.equal(f(ref_infos_with_cb, summary_container), summary, "callback returns summary");
    assert.equal(summary_container.text(), summary, "summary in container");


});
